# moved from tie_ml.modules.feature_extractor.tokenization

from dataclasses import dataclass

from tie_ml_utils.helper import TupleLike


@dataclass(slots=True)
class SentenceTokenization(TupleLike):
    """
    Sentence tokenized into sub tokens. Cannot be empty.
    """

    sub_token_ids: tuple[int, ...]
    tokens_start_mask: tuple[bool, ...]

    def __len__(self):
        return len(self.sub_token_ids)

    def __post_init__(self):
        if not len(self):
            raise ValueError('Empty tokenization is not allowed!')


@dataclass(slots=True)
class ContextualizedSentenceTokenization(SentenceTokenization):
    """
    Tokenization with added surrounding context.
    """

    main_sub_tokens_mask: tuple[bool, ...]


@dataclass(slots=True)
class TokenizedNodeChunk(ContextualizedSentenceTokenization):
    """
    Chunk of a document text. Can include text from multiple nodes, node_ids specify node id for each sentence.
    """

    node_ids: tuple[int, ...]
    sentence_ids: tuple[int, ...]
    shift: int

    def __post_init__(self):
        super(TokenizedNodeChunk, self).__post_init__()
        if not len(self.sentence_ids):
            raise ValueError('There should be at least one sentence in tokenized node!')
        if len(self.node_ids) != len(self.sentence_ids):
            raise ValueError('Length mismatch for node_ids and sentences!')


@dataclass(slots=True)
class TokenizedNodeChunkWithSentenceInfo(TokenizedNodeChunk):
    sentence_lengths: list[int]

    def __post_init__(self):
        super(TokenizedNodeChunk, self).__post_init__()
        if len(self.sentence_lengths) != len(self.sentence_ids):
            raise ValueError('Length mismatch for sentence_lengths and sentences!')



import json
from pathlib import Path
from typing import Any, Callable, Optional

from torch import Tensor

from tp_interfaces.serializable import DirectorySerializableMixin
from .abstract import AbstractPooler, choose


class AggregatePooler(AbstractPooler, DirectorySerializableMixin):

    def __init__(
            self,
            from_index: Optional[int] = None,
            to_index: Optional[int] = None,
            aggregate_function: Callable[[Tensor, int], Tensor] = None
    ):
        DirectorySerializableMixin.__init__(self)

        if aggregate_function is None:
            raise ValueError('No aggregate function specified!')

        self._from_index = from_index
        self._to_index = to_index
        self._aggregate_function = aggregate_function

    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        selected_elements = choose(tensor, self._from_index, self._to_index, dim=dim)
        return self._aggregate_function(selected_elements, dim)

    @classmethod
    def _serializer(cls, obj) -> tuple[Callable[[Any, Path], None], Callable[[Path], Any] | dict]:
        if isinstance(obj, Callable):
            return _save_callable, _load_callable
        return super()._serializer(obj)


def _save_callable(obj: Callable, path: Path):
    import inspect
    with path.open('w', encoding='utf-8') as f:
        json.dump({
            'module': inspect.getmodule(obj).__name__,
            'qualname': obj.__qualname__
        }, f, ensure_ascii=False, indent=2)


def _load_callable(path: Path) -> Callable:
    import importlib
    with path.open('r', encoding='utf-8') as f:
        config = json.load(f)
    module = importlib.import_module(config['module'])
    if '.' in config['qualname']:
        class_name, name = config['qualname'].rsplit('.', 1)
        module = getattr(module, class_name)
        config['qualname'] = name
    return getattr(module, config['qualname'])



# moved from tie_ml_base.torch_wrapper

from abc import ABCMeta, abstractmethod
from pathlib import Path
from typing import Any, AsyncContextManager

import torch
from torch.nn import Module
from typing_extensions import Self

from tp_interfaces.serializable import DirectorySerializableMixin
from tp_interfaces.serializable.pickle_mixin import PicklableMixin

DEFAULT_PREFERRED_DEVICE = 'cuda' if torch.cuda.is_available() else 'cpu'


class TorchModule(Module, AsyncContextManager, metaclass=ABCMeta):

    def __init__(self, preferred_device: str | torch.device = None):
        super().__init__()
        self._preferred_device = preferred_device or DEFAULT_PREFERRED_DEVICE

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def device(self) -> torch.device:
        try:
            return next(self.parameters()).device
        except StopIteration:
            return torch.device(self._preferred_device)

    @property
    def preferred_device(self) -> torch.device:
        return torch.device(self._preferred_device)

    @abstractmethod
    def forward(self, *args, **kwargs) -> Any:
        pass


class PicklableTorchModule(TorchModule, PicklableMixin, metaclass=ABCMeta):

    def save(self, path: Path, *, rewrite: bool = False) -> None:
        previous_device = self.device
        self.cpu()
        super(PicklableTorchModule, self).save(path, rewrite=rewrite)
        self.to(device=torch.device(previous_device))


class DirectorySerializableTorchModule(TorchModule, DirectorySerializableMixin, metaclass=ABCMeta):
    """
    Base class for torch modules with implementation for serialization and deserialization in directory.
    It uses base strategy for object storing, but also saves weights

    @note this mixin should be initialized before torch model is built
    """

    def __init__(self):
        TorchModule.__init__(self)
        DirectorySerializableMixin.__init__(self)

    @classmethod
    def _load_from_directory(cls, directory_path: Path, info: dict) -> Self:
        model = super()._load_from_directory(directory_path, info)
        if not isinstance(model, DirectorySerializableTorchModule):
            raise ValueError
        model.load_state_dict(torch.load(directory_path / 'weights.pt'), strict=False)
        return model

    def _save_to_directory(self, directory_path: Path) -> None:
        super()._save_to_directory(directory_path)

        state_dict = self.state_dict()
        for name, (value, _, _) in self._serializable.items():
            if isinstance(value, TorchModule):  # assume it was saved with parameters
                for param in value.state_dict():
                    state_dict.pop(f'_{name}.{param}', None)

        torch.save(state_dict, directory_path / 'weights.pt')



# moved from tie_ml.modules.feature_extractor.helper

import io
import pickle
from hashlib import md5
from typing import Any

import torch
from torch import BoolTensor, Tensor
from torch.nn import Parameter

from tie_ml_utils.helper import fill_roll


def parameter_hash(param: Parameter) -> str:
    buffer = io.BytesIO()
    pickle.dump(param.data.tolist(), buffer)
    buffer.seek(0)
    return md5(buffer.read()).hexdigest()


def extract_padded_token_images(
        sub_token_representations: Tensor,
        token_masks: BoolTensor,
        padding_masks: BoolTensor,
        *,
        pad_value: Any = 0.0
) -> Tensor:
    """Converts sub token representation into token images.
    Token image - token's sub token representations, concatenated into 'image'.

    For example, token with the following sub token representations: [1, 2], [3, 4], [5, 6]
    forms a token image:
    [
        [1, 2],
        [3, 4],
        [5, 6]
    ]

    :param: sub_token_representations - tensor of shape (B, ST, F)
    :param: token_masks - tensor of shape (B, ST) that indicates starts of tokens
    :param: padding_masks - tensor of shape (B, ST) that indicates non-pad sub tokens
    :param: pad_value - padding value for resulting images
    :return: tensor of shape (B, T, I, F)

    B - batch size
    ST - maximum length of sequences in sub tokens
    F - number of features
    T - maximum length of sequences in tokens
    I - maximum number of sub tokens in tokens
    """

    device = sub_token_representations.device
    batch_size, _, n_features = sub_token_representations.shape

    # calculate number of tokens in each batch sequence

    n_tokens = token_masks.sum(dim=-1)
    pad_sequences_to_size = n_tokens.max()

    # calculate how much padding tokens should be skipped for each batch sequence

    padding_tokens = torch.cumsum(pad_sequences_to_size - n_tokens, dim=-1)
    padding_tokens_skip = fill_roll(padding_tokens, shift=1, fill=0)

    # calculate image sizes (number of sub tokens in each token)

    cumulative_tokens_count = torch.cumsum(token_masks.view(-1), dim=-1)
    _, image_sizes = torch.unique_consecutive(cumulative_tokens_count[padding_masks.view(-1)], return_counts=True)

    # reserve space for padded images

    pad_images_to_size = image_sizes.max()
    padded_shape = (batch_size, pad_sequences_to_size, pad_images_to_size, n_features)
    padded_token_images = torch.full(padded_shape, fill_value=pad_value, dtype=sub_token_representations.dtype, device=device)

    # calculate shifts in dimensions for each sub token

    # FIXME: suboptimal
    image_dim_shift = torch.cat(list(map(torch.arange, image_sizes)))

    sequence_dim_shift = (cumulative_tokens_count - 1).view(batch_size, -1) + padding_tokens_skip.view(batch_size, 1)
    total_shift = sequence_dim_shift[padding_masks] * pad_images_to_size.to(device) + image_dim_shift.to(device)

    # fill feature representations into images

    raveled_padded_token_images = padded_token_images.view(-1, n_features)
    raveled_sub_token_representations = sub_token_representations[padding_masks]
    raveled_padded_token_images[total_shift] = raveled_sub_token_representations

    return padded_token_images



import gc
import inspect
import logging
from functools import wraps
from os import environ
from random import shuffle
from time import sleep
from typing import Any, Callable, Optional, Tuple

import torch
from pynvml import NVMLError, nvmlDeviceGetHandleByIndex, nvmlDeviceGetMemoryInfo, nvmlDeviceGetUtilizationRates, nvmlInit
from torch.nn import Module

from talisman_tools.helper.log_tools import warn_once
from tie_ml_utils.env import get_gpu_oom_repeat_limit, get_gpu_oom_timeout, get_gpu_utilization_limit
from tie_ml_utils.torch_wrapper import TorchModule

logger = logging.getLogger(__name__)
try:
    nvmlInit()
    nvml_loaded = True
except NVMLError as nvml_error:
    logger.warning(f'Failed to load NVML library: {nvml_error}')
    nvml_loaded = False


def model_bytes(model: TorchModule) -> int:
    """Returns model size in bytes."""
    param_size = 0
    for param in model.parameters():
        param_size += param.nelement() * param.element_size()

    buffer_size = 0
    for buffer in model.buffers():
        buffer_size += buffer.nelement() * buffer.element_size()

    return param_size + buffer_size


def get_real_device_id(torch_device_id: int) -> int:
    visible_devices = environ.get('CUDA_VISIBLE_DEVICES')
    if visible_devices is None:
        return torch_device_id

    visible_devices = list(map(int, visible_devices.split(',')))
    real_device_ids = list(range(len(visible_devices)))
    device_mapping = dict(zip(real_device_ids, visible_devices))

    return device_mapping[torch_device_id]


def get_cuda_device_with_most_memory() -> Tuple[Optional[int], int]:
    """Returns CUDA device with most memory, as well as the available memory on the device in bytes."""
    most_unoccupied = 0
    target_gpu_id = None

    if not nvml_loaded:
        warn_once(logger, 'NVML library is not loaded, no information about global state of GPU is available! '
                          'This may lead to suboptimal choice of the GPU.')

    devices = list(range(torch.cuda.device_count()))
    shuffle(devices)
    for device_id in devices:
        total_bytes = torch.cuda.get_device_properties(device_id).total_memory
        allocated_bytes = torch.cuda.memory_allocated(device_id)
        reserved_bytes = torch.cuda.memory_reserved(device_id)

        if nvml_loaded:
            nvml_device_id = get_real_device_id(device_id)
            device_handle = nvmlDeviceGetHandleByIndex(nvml_device_id)

            gpu_utilization = nvmlDeviceGetUtilizationRates(device_handle).gpu / 100.0
            if gpu_utilization > get_gpu_utilization_limit():
                continue

            used_bytes = nvmlDeviceGetMemoryInfo(device_handle).used
            cached_bytes = reserved_bytes - allocated_bytes  # torch caches memory without using it
            available_bytes = total_bytes - (used_bytes - cached_bytes)  # used memory includes cached torch memory
        else:
            available_bytes = total_bytes - allocated_bytes  # does not include information about other processes

        if available_bytes > most_unoccupied:
            most_unoccupied = available_bytes
            target_gpu_id = device_id

    return target_gpu_id, most_unoccupied


def torch_device_handler(f: Callable):
    """Attempts to call the function with the most vacant GPU as a default CUDA device. Repeats until GPU_OOM_REPEAT_LIMIT is reached."""

    def wrapper(*args, **kwargs):
        timeout = get_gpu_oom_timeout()
        repeats = 0

        while True:
            repeats += 1
            try:
                device_id, _ = get_cuda_device_with_most_memory()
                if device_id is None:
                    logger.warning(f'No device is available! Repeating after {timeout} seconds!')
                    sleep(timeout)
                    continue

                with torch.cuda.device(device_id):
                    return f(*args, **kwargs)
            except RuntimeError as e:
                if repeats > get_gpu_oom_repeat_limit():
                    raise RuntimeError('GPU_OOM_REPEAT_LIMIT is reached! Consider either raising the limit or the GPU_OOM_TIMEOUT!')

                logger.warning(f'Caught runtime error while calling function {f}. Repeating after {timeout} seconds!', exc_info=e)
                sleep(timeout)

    if not torch.cuda.is_available():
        logger.warning(f'No CUDA devices available on your system!')
        return f

    return wrapper


def cuda_handler_cleanup(self: TorchModule, e: RuntimeError, f: Callable[[Module, Any], Any]):
    logger.error(f'Caught runtime error while calling async function {f}. Moving {self.__class__.__name__} to CPU!', exc_info=e)
    del e
    self.cpu()
    gc.collect()
    torch.cuda.empty_cache()


def cuda_handler(f: Callable[[Module, Any], Any]):
    """Attempts to place the model (the first argument of the function f) on GPU before calling the function.
    See `try_place_on_cuda` for GPU placement criteria."""

    @wraps(f)
    async def async_wrapper(self: TorchModule, *args, **kwargs):
        if self.preferred_device == torch.device('cpu'):
            warn_once(logger, f'{self.__class__.__name__} preferred device is set to CPU, keeping it on CPU!')
            return await f(self, *args, **kwargs)
        try:
            try_place_on_cuda(self)
            return await f(self, *args, **kwargs)
        except RuntimeError as e:
            cuda_handler_cleanup(self, e, f)
            return await f(self, *args, **kwargs)

    @wraps(f)
    def sync_wrapper(self: TorchModule, *args, **kwargs):
        if self.preferred_device == torch.device('cpu'):
            warn_once(logger, f'{self.__class__.__name__} preferred device is set to CPU, keeping it on CPU!')
            return f(self, *args, **kwargs)
        try:
            try_place_on_cuda(self)
            return f(self, *args, **kwargs)
        except RuntimeError as e:
            cuda_handler_cleanup(self, e, f)
            return f(self, *args, **kwargs)

    if inspect.iscoroutinefunction(f) or inspect.isasyncgenfunction(f):
        return async_wrapper
    return sync_wrapper


def try_place_on_cuda(model: TorchModule):
    """Places model on a GPU if the following criteria are met:
    1. There are GPU devices that are visible.
    2. Model is not already on GPU.
    3. There are enough memory to fit model on GPU.

    If any of these criteria are not satisfied or RuntimeError occurred during model placement on GPU,
    model will be placed on CPU.
    """
    if not torch.cuda.is_available():
        return  # no cuda is available

    if model.device.type == 'cuda':
        return  # model is already on cuda

    device_id, available_memory = get_cuda_device_with_most_memory()
    if available_memory < model_bytes(model) or device_id is None:
        return  # not enough memory available on GPU

    logger.warning(f'Found GPU devices available: {torch.cuda.device_count()}. Placing {model.__class__.__name__} on GPU!')
    try:
        device = torch.device(f'cuda:{device_id}')
        model.to(device)
        logger.warning(f'{model.__class__.__name__} is successfully placed on {model.device}.')
    except RuntimeError as e:
        logger.error(f'Caught runtime error while placing {model.__class__.__name__} on GPU. '
                     f'Moving {model.__class__.__name__} to CPU!', exc_info=e)
        del e

        model.cpu()
        gc.collect()
        torch.cuda.empty_cache()



from torch import Tensor

from tp_interfaces.serializable import DirectorySerializableMixin
from .abstract import AbstractPooler


class SelectPooler(AbstractPooler, DirectorySerializableMixin):

    def __init__(self, index: int):
        DirectorySerializableMixin.__init__(self)
        self._index = index

    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        if dim == 0 or dim == -len(tensor.shape):
            return tensor[self._index]
        return tensor.swapdims(0, dim)[[self._index]].swapdims(0, dim).squeeze(dim)



from functools import wraps
from typing import Any, Callable

import torch
from torch import amp

from tie_ml_utils.env import get_amp_optimize
from tie_ml_utils.torch_wrapper import PicklableTorchModule


def optimize(f: Callable[[PicklableTorchModule, Any], Any]):
    @wraps(f)
    @torch.no_grad()
    def amp_wrapper(self: PicklableTorchModule, *args, **kwargs):
        if self.device.type == 'cuda' and get_amp_optimize():  # for some reason AMP with CPU is very slow
            with amp.autocast(device_type='cuda'):
                return f(self, *args, **kwargs)
        return f(self, *args, **kwargs)

    return amp_wrapper



import math
from typing import Tuple

import torch
from torch import Tensor

from tie_ml_utils.torch_wrapper import DirectorySerializableTorchModule
from .abstract import AbstractPooler


class AttentionPooler(AbstractPooler, DirectorySerializableTorchModule):
    def __init__(self, features: int, dropout_rate: float):
        DirectorySerializableTorchModule.__init__(self)
        self._attention = torch.nn.Linear(in_features=features, out_features=1)
        self._dropout = torch.nn.Dropout(dropout_rate)
        self._scaling_factor = math.sqrt(features)

    def _validate_dim(self, shape: Tuple[int, ...], dim: int):
        if dim == -1 or dim == len(shape) - 1:
            raise ValueError(f'{self.__class__.__name__} cannot pool feature dimension!')
        if len(shape) < 3:
            raise ValueError(f'{self.__class__.__name__} requires at least 3 dimensions but {len(shape)} were given!')
        if shape[-1] != self._attention.in_features:
            raise ValueError(f'Expected feature dimension to be {self._attention.in_features} but got {shape[-1]}!')

    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        # wrapper around call that the parent class needs to have
        return self.__call__(tensor, dim=dim)

    def forward(self, tensor: Tensor, *, dim: int) -> Tensor:
        """
        Pools `tensor` by applying scaled dot-product attention and dropout.
        Args:
            tensor: a (3+)-dimensional tensor of shape N1, ..., Nm, F where F is an embedding dimension
            dim: pooled dimension, last dimension cannot be pooled

        Returns:
            tokens: tensor of the same shape as input tensor but with pooling dimension missing
        """
        orig_dims = tensor.shape  # shape N1, ..., Nm, F
        self._validate_dim(orig_dims, dim)

        # move pooled dimension at -2
        tensor = torch.swapdims(tensor, dim, -2)  # shape N_1, ..., N_dim-1, N_m, N_dim+1, ..., N_m-1, N_dim, F

        attn_scores = self._attn_scores(tensor)  # shape N_1, ..., N_dim-1, N_m, N_dim+1, ..., N_m-1, N_dim
        attn_scores = attn_scores.unsqueeze(-2)  # shape N_1, ..., N_dim-1, N_m, N_dim+1, ..., N_m-1, 1, N_dim

        # N = N_1 * ... * N_dim-1 * N_m * N_dim+1 * ... * N_m-1
        # BMM operation on (N, 1, N_dim) x (N, N_dim, F) -> (N, 1, F)
        unravelled_dims = attn_scores.shape[:-2]
        pooled_dim = attn_scores.shape[-1]
        result = torch.bmm(attn_scores.reshape(-1, 1, pooled_dim), tensor.reshape(-1, pooled_dim, self._attention.in_features))

        # restore original dims
        result = result.reshape(*unravelled_dims, 1, self._attention.in_features)  # shape N1, ..., N_dim-1, N_m, N_dim+1, ..., N_m-1, 1, F
        result = torch.swapdims(result, -2, dim)  # shape N1, ..., N_dim-1, 1, N_dim+1, ..., N_m, F
        return result.squeeze(dim)  # remove pooled dim, shape N1, ..., N_dim-1, N_dim+1, ..., N_m, F

    def _attn_scores(self, example_representations: Tensor):
        # mask with 0 at positions belonging to the words and -inf elsewhere to mask out padding
        attn_mask = example_representations.sum(dim=-1) != 0
        not_empty_examples_mask = attn_mask.sum(dim=-1) != 0

        attn_mask = torch.log(attn_mask.to(torch.float))
        attn_mask[~not_empty_examples_mask] = 0  # for empty examples we will get constant scores that will not affect gradients

        attn_scores = self._attention(self._dropout(example_representations))  # [..., 1], removed transposition
        attn_scores = attn_mask + attn_scores.squeeze(dim=-1)  # ONLY squeeze last dim - otherwise error with len 1 dims

        # apply scaling factor as in https://arxiv.org/pdf/1706.03762v5.pdf to preserve variance 1
        return torch.softmax(attn_scores / self._scaling_factor, dim=-1)



# moved from tie_ml.modules.feature_extractor.batching

import dataclasses
from dataclasses import dataclass
from functools import partial
from itertools import chain, starmap
from typing import Any, Callable, Generic, Iterable, Optional, Sequence, TypeVar

import numpy as np
import torch
from torch import BoolTensor, LongTensor, Tensor
from torch.nn.utils.rnn import pad_sequence
from typing_extensions import Self

from tie_datamodel.datamodel.node.text import TIETextNode
from tie_ml_utils.helper import TensorStorage, fill_roll_np
from .tokenization import (
    SentenceTokenization, TokenizedNodeChunk, TokenizedNodeChunkWithSentenceInfo
)

_Tokenization = TypeVar('_Tokenization', bound=SentenceTokenization)


@dataclass(slots=True)
class DocumentExample(Generic[_Tokenization]):
    doc_id: int
    doc_total_chunks: int
    chunk_id: int
    tokenization: _Tokenization

    def __len__(self):
        return len(self.tokenization)

    @classmethod
    def from_nodes(
            cls,
            nodes: Iterable[Sequence[TIETextNode]],
            chunker: Callable[[Sequence[TIETextNode]], Iterable[TokenizedNodeChunk]]
    ) -> Iterable[Self]:
        """Converts nodes into examples. Note that each node can be split into several text chunks and,
        subsequently, into several examples.
        """
        yield from chain.from_iterable(starmap(
            partial(cls._process_doc_contents, chunker=chunker),
            enumerate(nodes)
        ))

    @classmethod
    def _process_doc_contents(
            cls,
            doc_id: int,
            nodes: Sequence[TIETextNode],
            chunker: Callable[[Sequence[TIETextNode]], Iterable[TokenizedNodeChunk]]
    ) -> tuple[Self, ...]:
        sentence_lengths = [[len(sentence) for sentence in node.sentences] for node in nodes]
        chunks = tuple(chunker(nodes))

        def example_builder(example_id: int, example: TokenizedNodeChunk) -> DocumentExample:
            return cls(
                doc_id=doc_id,
                doc_total_chunks=len(chunks),
                chunk_id=example_id,
                tokenization=TokenizedNodeChunkWithSentenceInfo(
                    **dataclasses.asdict(example),
                    sentence_lengths=[
                        sentence_lengths[node_id][sentence_id]
                        for node_id, sentence_id in zip(example.node_ids, example.sentence_ids)
                    ]
                )
            )

        return tuple(starmap(example_builder, enumerate(chunks)))


_Any = TypeVar('_Any')


def sort_intervals(iterable: Iterable[_Any], *, interval_length: int, key: Callable[[_Any], Any]) -> Iterable[_Any]:
    intervals = group_to_batches(iterable, batch_size=interval_length)
    for interval in intervals:
        yield from sorted(interval, key=key)


def group_to_batches(iterable: Iterable[_Any], *, batch_size: int, length_sort_batches: int = 0) -> Iterable[list[_Any]]:
    batch = []
    if length_sort_batches > 0:
        iterable = sort_intervals(iterable, interval_length=batch_size * length_sort_batches, key=len)
    for example in iterable:
        if len(batch) == batch_size:
            yield batch
            batch = []

        batch.append(example)

    if len(batch):
        yield batch


@dataclass(slots=True)
class BatchedExamples(TensorStorage):
    """
    Dataclass for communicating between modules.
    Information about main sentences and context is populated by feature extractors.
    Label scores are added by sequence taggers.

    BatchedExample can be converted to label sequence by decoders.

    B - batch size
    T - sequence length in tokens (without context)
    CT - sequence length in tokens (with context)
    SCT - sequence length in sub tokens (with context)
    F - size of representation (number of features)
    N - number of possible labels
    """

    doc_ids: list[int]
    """Document ids that examples sentences originated from."""

    doc_total_chunks: list[int]
    """Total number of chunks for each document."""

    chunk_ids: list[int]
    """Chunk id of each example."""

    sentence_ids: list[list[int]]
    """Ids of sentences inside nodes."""

    sentence_lengths: list[list[int]]
    """Lengths of examples sentences."""

    sentence_starts: list[list[int]]
    """Sentences start positions in examples. Is, practically, [0, cumsum(sentence_lengths)[:-1]]"""

    node_ids: list[list[int]]
    """Node ids that examples sentences originated from."""

    shift: list[int]
    """Starting position in original segmentation of the first sentence."""

    # Sub token level

    sub_token_ids: Optional[LongTensor] = None
    """Segmentation of examples used by feature extractor. Shape: (B, SCT)."""

    sub_token_representations: Optional[Tensor] = None
    """Sub token level representation of sentences generated by feature extractor. Shape: (B, SCT, F)."""

    sub_token_padding_mask: Optional[BoolTensor] = None
    """Mask that indicates padding for sub tokens. Shape: (B, ST)"""

    main_sub_tokens_mask: Optional[BoolTensor] = None
    """Mask that separates context from main sentences in sub tokens. Shape: (B, SCT)."""

    tokens_start_mask: Optional[BoolTensor] = None
    """Mask that indicates the start of each token from original document segmentation (without context). Shape: (B, SCT)."""

    # Token level

    token_representations: Optional[Tensor] = None
    """Token level representation of sentences generated by feature extractor. Shape: (B, CT, F)."""

    token_padding_mask: Optional[BoolTensor] = None
    """Mask that indicates padding for tokens. Shape: (B, CT)"""

    main_tokens_mask: Optional[BoolTensor] = None
    """Mask that separates context from main sentences in tokens. Shape: (B, CT)."""

    # Label level

    main_tokens_representations: Optional[Tensor] = None
    """Token level representations without context tokens. Shape: (B, T, F)."""

    label_scores: Optional[Tensor] = None
    """Label scores for each token generated by the sequence tagger. Shape: (B, T, N)."""

    label_padding_mask: Optional[BoolTensor] = None
    """Mask that indicates label paddings. Shape: (B, T)."""

    def replace(
            self,
            *,
            drop_others: bool = False,
            drop_fields: Iterable[str] = tuple(),
            keep_fields: Iterable[str] = tuple(),
            **changes
    ) -> Self:

        new_examples = dataclasses.replace(self, **changes)

        to_drop = set()
        if drop_others:
            field_names = {field.name for field in dataclasses.fields(new_examples)}
            to_drop = field_names.difference(chain(changes.keys(), keep_fields))
        to_drop.update(drop_fields)
        return new_examples.drop(*to_drop)

    def drop(self, *attributes) -> Self:
        if not len(attributes):
            return self
        changes = dict.fromkeys(attributes, None)
        return dataclasses.replace(self, **changes)

    @classmethod
    def from_examples(cls, examples: Iterable[DocumentExample[TokenizedNodeChunkWithSentenceInfo]], *, pad_sub_token_id: int) -> Self:
        doc_ids: list[int] = []
        doc_total_chunks: list[int] = []
        chunk_ids: list[int] = []
        sentence_ids: list[list[int]] = []
        sentence_lengths: list[list[int]] = []
        sentence_starts: list[list[int]] = []
        node_ids: list[list[int]] = []
        shift: list[int] = []
        sub_token_ids: list[LongTensor] = []
        sub_token_padding_mask: list[BoolTensor] = []
        main_sub_tokens_mask: list[BoolTensor] = []
        tokens_start_mask: list[BoolTensor] = []

        def long_convert(container: Sequence) -> LongTensor:
            if torch.is_tensor(container):
                container: Tensor
                return container.long()
            return torch.tensor(container, dtype=torch.long).long()

        def bool_convert(container: Sequence) -> BoolTensor:
            if torch.is_tensor(container):
                container: Tensor
                return container.bool()
            return torch.tensor(container, dtype=torch.bool).bool()

        # extracting data from examples and converting into tensors

        for example in examples:
            doc_ids.append(example.doc_id)
            doc_total_chunks.append(example.doc_total_chunks)
            chunk_ids.append(example.chunk_id)
            sentence_ids.append(list(example.tokenization.sentence_ids))
            sentence_lengths.append(example.tokenization.sentence_lengths)
            sentence_starts.append(list(fill_roll_np(np.cumsum(np.array(example.tokenization.sentence_lengths)), shift=1, fill=0)))
            node_ids.append(list(example.tokenization.node_ids))
            shift.append(example.tokenization.shift)

            sub_token_ids_tensor = long_convert(example.tokenization.sub_token_ids)
            sub_token_padding_mask_tensor = torch.ones_like(sub_token_ids_tensor, dtype=torch.bool).bool()
            main_sub_tokens_mask_tensor = bool_convert(example.tokenization.main_sub_tokens_mask)
            tokens_start_mask_tensor = bool_convert(example.tokenization.tokens_start_mask)

            sub_token_ids.append(sub_token_ids_tensor)
            sub_token_padding_mask.append(sub_token_padding_mask_tensor)
            main_sub_tokens_mask.append(main_sub_tokens_mask_tensor)
            tokens_start_mask.append(tokens_start_mask_tensor)

        # combining collected tensors into one

        pad_collator = partial(pad_sequence, batch_first=True)

        sub_token_ids: LongTensor = pad_collator(sub_token_ids, padding_value=pad_sub_token_id).long()
        sub_token_padding_mask: BoolTensor = pad_collator(sub_token_padding_mask, padding_value=False).bool()
        main_sub_tokens_mask: BoolTensor = pad_collator(main_sub_tokens_mask, padding_value=False).bool()
        tokens_start_mask: BoolTensor = pad_collator(tokens_start_mask, padding_value=False).bool()

        result = cls(
            doc_ids=doc_ids,
            doc_total_chunks=doc_total_chunks,
            chunk_ids=chunk_ids,
            sentence_ids=sentence_ids,
            sentence_lengths=sentence_lengths,
            sentence_starts=sentence_starts,
            node_ids=node_ids,
            shift=shift,
            sub_token_ids=sub_token_ids,
            sub_token_padding_mask=sub_token_padding_mask,
            main_sub_tokens_mask=main_sub_tokens_mask,
            tokens_start_mask=tokens_start_mask
        )

        # all sub token masks are present, so we can fill in token masks
        return add_token_masks(result)

    @classmethod
    def from_nodes(
            cls,
            nodes: Iterable[Sequence[TIETextNode]],
            chunker: Callable[[Sequence[TIETextNode]], Iterable[TokenizedNodeChunk]],
            *,
            batch_size: int,
            pad_sub_token_id: int,
            group_by_length: bool = False
    ) -> Iterable[Self]:

        examples_collator = partial(BatchedExamples.from_examples, pad_sub_token_id=pad_sub_token_id)
        examples = DocumentExample.from_nodes(nodes, chunker)
        yield from map(examples_collator, group_to_batches(examples, batch_size=batch_size, length_sort_batches=int(group_by_length) * 10))

    def get_length_in_tokens(self):
        """
        NB: only functions when chunker overlap is 1
        """
        token_counts = self.main_tokens_mask.sum(dim=-1)
        if len(token_counts) > 1:
            return sum(token_counts)
        else:
            return token_counts.item()

    def get_full_batch_no_overlap(self):
        """
        NB: this only works if chunker overlap is set to 1. Otherwise this will fail.
        """
        return self.token_representations[self.main_tokens_mask]


def add_token_masks(batch: BatchedExamples) -> BatchedExamples:
    """Add token masks based on sub token masks"""

    token_padding_mask = []
    main_tokens_mask = []

    masks_iterator = zip(
        batch.tokens_start_mask,
        batch.main_sub_tokens_mask,
        batch.sub_token_padding_mask
    )

    for tokens_start_mask, main_sub_tokens_mask, sub_token_padding_mask in masks_iterator:
        sequence_length = tokens_start_mask[sub_token_padding_mask].sum()
        token_padding_mask.append(torch.ones(sequence_length, dtype=torch.bool))

        main_tokens_mask.append(main_sub_tokens_mask[sub_token_padding_mask & tokens_start_mask])  # without padding and for tokens

    # pad sequences to equal length
    pad_collator = partial(pad_sequence, batch_first=True)

    token_padding_mask = pad_collator(token_padding_mask, padding_value=False).bool()
    main_tokens_mask = pad_collator(main_tokens_mask, padding_value=False).bool()

    return batch.replace(token_padding_mask=token_padding_mask, main_tokens_mask=main_tokens_mask)


def repad(representations: Tensor, mask: BoolTensor) -> tuple[Tensor, BoolTensor]:
    """Apply mask and repad tensor.

    :param: tensor - tensor with shape (B, S, F)
    :param: mask - tensor with the shape (B, S)
    :return: new tensor and padding mask

    B - batch size
    S - sequence length
    F - number of features
    """

    device = representations.device
    _, _, num_features = representations.shape

    all_tensor_rows: list[Tensor] = []
    paddings: list[BoolTensor] = []
    for tensor_row, mask_row in zip(representations, mask):
        masked_row = tensor_row[mask_row].reshape(-1, num_features)
        padding: BoolTensor = torch.ones(len(masked_row), dtype=torch.bool, device=device).bool()

        all_tensor_rows.append(masked_row)
        paddings.append(padding)

    # pad sequences to equal length
    pad_collator = partial(pad_sequence, batch_first=True)

    return pad_collator(all_tensor_rows), pad_collator(paddings, padding_value=False)


def get_representations_for_main_tokens(
        examples: BatchedExamples,
        *,
        drop_others: bool = True,
        keep_fields: Iterable[str] = tuple(),
        drop_fields: Iterable[str] = tuple()
) -> BatchedExamples:
    main_tokens_representation, label_padding_mask = repad(examples.token_representations, examples.main_tokens_mask)
    return examples.replace(
        main_tokens_representations=main_tokens_representation,
        label_padding_mask=label_padding_mask,
        drop_others=drop_others,
        keep_fields=keep_fields,
        drop_fields=drop_fields
    )



# moved from tie_ml.modules.sequence_tagger.decoder.labelling_strategy

import logging
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from copy import copy
from enum import Enum
from itertools import chain, starmap
from typing import Iterable, Iterator, NamedTuple, Sequence

import torch
from tdm.abstract.datamodel import AbstractFact, FactStatus
from tdm.datamodel.facts import MentionFact
from tdm.datamodel.mentions import TextNodeMention
from tdm.utils import MentionedFactsFactory
from torch import LongTensor

from tie_datamodel.datamodel.node.text import TIETextNode
from tie_datamodel.datamodel.span import Span

logger = logging.getLogger(__name__)


class LabellingStrategyType(str, Enum):
    BIO = 'BIO'
    IOB = 'IOB'


class MentionSignature(NamedTuple):
    start: int
    end: int


class AbstractLabellingStrategy(metaclass=ABCMeta):
    """
    An abstract utility class that defines an interface of functions that:
    1) encode text into a plain sequence of labels (basically for training/evaluation purposes, not inference)
    2) decode label_ids according to the selected labelling strategy

    Note that labels are a combination of entity type and label prefix except non-entity label.
    Also keep in mind that entity type is not the same as the fact type.
    """

    def __init__(self, ent_labels: set[str], outside_ent_label: str = 'O'):
        """
        :param ent_labels: label prefixes denoting entities
        :param outside_ent_label: non-entity label
        """
        self._ent_labels = copy(ent_labels)
        self._outside_ent_label = outside_ent_label

    def get_possible_categories(self, ent_types: set[str]) -> set[str]:
        """
        Returns possible labels based on label prefixes and entity types.
        """
        return {
            f'{label}-{e_type}'
            for label in self._ent_labels for e_type in ent_types
        }.union((self._outside_ent_label,))

    @property
    @abstractmethod
    def type(self) -> LabellingStrategyType:
        """
        Type of used labelling strategy.
        """
        pass

    @property
    def outside_ent_label(self) -> str:
        """
        Non-entity label.
        """
        return self._outside_ent_label

    @abstractmethod
    def encode_labels(self, node: TIETextNode, mentions: Iterable[MentionFact]) -> list[str]:
        """
        Encodes node text into a list of labels (one label per span).
        Note that during encoding, the entity type and the fact type are the same
        and only labels based on the AtomValueFact type IDs are used.

        :param node: given document node with text
        :param mentions: sequence of mention facts
        """
        pass

    @abstractmethod
    def decode_signatures(
            self,
            label_ids: LongTensor,
            label_ids_mapping: Sequence[str],
            *,
            ignore_label_id: int = -100
    ) -> dict[str, list[MentionSignature]]:
        """
        Decodes label_ids into labels and returns a dict of extracted model entity types with the corresponding mentions.

        :param label_ids: given sequence of labels ids to decode
        :param label_ids_mapping: sequence of used labels
        :param ignore_label_id: ID of ignored label
        """
        pass

    @abstractmethod
    def decode_label_ids(
            self,
            node: TIETextNode,
            label_ids: LongTensor,
            *,
            type_mapping: dict[str, set[MentionedFactsFactory]],
            label_ids_mapping: Sequence[str],
            precomputed_tokens: Sequence[Span] | None = None
    ) -> Iterator[AbstractFact]:
        """
        Decodes label_ids into labels and returns a set of extracted facts.

        :param node: original document node with text
        :param label_ids: given sequence of labels ids to decode
        :param domain: used domain
        :param type_mapping: mapping from entity types to the corresponding facts factory
        :param label_ids_mapping: sequence of used labels
        :param precomputed_tokens: precomputed tokens from the given node
        """
        pass


def _greedy_resolve_intersections(entities: Iterator[MentionFact]) -> Iterator[MentionFact]:
    try:
        prev = next(entities)
    except StopIteration:
        return
    yield prev
    for current in entities:
        prev_span = Span(prev.mention.start, prev.mention.end)
        current_span = Span(current.mention.start, current.mention.end)
        if current_span.intersects(prev_span):
            def format_mention(fact: MentionFact) -> str:
                mention = fact.mention
                mention_text = mention.node.content[mention.start:mention.end]
                return f'<{fact.value.str_type_id} at [{mention.start}, {mention.end}] in {mention.node_id}: "{mention_text}">'

            logger.warning(f"Intersecting entities mentions:\n"
                           f"\t{format_mention(prev)};\n"
                           f"\t{format_mention(current)}.\n"
                           f"Only first one will be used")
            continue
        yield current
        prev = current


def _match_to_tokens(tokens: Iterator[Span], entities: Iterator[MentionFact]) -> Iterator[tuple[int, int, str]]:
    try:
        current_entity = next(entities)
    except StopIteration:
        return  # no entities

    current_start: int | None = None
    token_idx = 0
    for token_idx, token in enumerate(tokens):
        while current_entity.mention.end <= token.start:
            if current_start is not None:
                yield current_start, token_idx, current_entity.value.str_type_id
                current_start = None
            try:
                current_entity = next(entities)
            except StopIteration:
                return
        if current_start is None and token.start >= current_entity.mention.start and token.end <= current_entity.mention.end:
            current_start = token_idx
        # TODO: check token boundary intersection
    if current_start is not None:
        yield current_start, token_idx + 1, current_entity.value.str_type_id


class _BIOFactoryLabellingStrategy(AbstractLabellingStrategy, metaclass=ABCMeta):
    def __init__(self, start_symbol: str = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._start_symbol = start_symbol if start_symbol is not None else None

    def encode_labels(self, node: TIETextNode, mentions: Iterable[MentionFact]) -> list[str]:
        text_mentions = filter(lambda m: isinstance(m.mention, TextNodeMention), mentions)
        # sort for greedy intersection resolution
        sorted_text_mentions = sorted(text_mentions, key=lambda m: (m.mention.start, -m.mention.end))
        labels = [self._outside_ent_label] * sum(map(len, node.sentences))

        if not len(sorted_text_mentions):
            return labels

        for ent_start, ent_end, ent_label in _match_to_tokens(node.tokens, _greedy_resolve_intersections(iter(sorted_text_mentions))):
            last_label = labels[ent_start - 1]  # it may get the label on position -1 on first iteration (it will be `O`)
            entity_labels = self._encode_entity(ent_end - ent_start, ent_label, last_label)
            labels[ent_start:ent_end] = (f'{label}-{ent_label}' for label in entity_labels)

        return labels

    def _is_object(self, label: str) -> bool:
        return label != self._outside_ent_label

    @staticmethod
    def _get_object_type(label: str) -> str:
        return '-'.join(label.split('-')[1:])

    @staticmethod
    def _get_object_encoding(label: str) -> str:
        return label.split('-')[0]

    def decode_signatures(
            self,
            label_ids: LongTensor,
            label_ids_mapping: Sequence[str],
            *,
            ignore_label_id: int = -100
    ) -> dict[str, list[MentionSignature]]:

        mentions: list[int] = []
        fact_type = None
        compressed_ids, counts = torch.unique_consecutive(label_ids, return_counts=True)

        label_position = 0
        fact_signatures: dict[str, list[MentionSignature]] = defaultdict(list)
        for label_id, count in zip(compressed_ids, counts):
            next_label_position = label_position + count
            positions = tuple(range(label_position, next_label_position))
            label_position = next_label_position

            if label_id == ignore_label_id:
                continue

            label = label_ids_mapping[label_id.item()]

            encode = self._get_object_encoding(label)
            current_type = self._get_object_type(label)

            if encode == self._start_symbol:  # beginning of a new entity
                if fact_type is not None:
                    # finish current entity
                    fact_signatures[fact_type].append(MentionSignature(mentions[0], mentions[-1]))

                # several start labels in a row == series of one-span entities
                fact_signatures[current_type].extend(MentionSignature(position, position) for position in positions[:-1])

                # start new entity for last span
                mentions = [positions[-1]]
                fact_type = current_type
                continue

            if encode != self._outside_ent_label and encode != self._start_symbol:  # inside an entity
                if fact_type is not None and fact_type != current_type:
                    # finish current entity
                    fact_signatures[fact_type].append(MentionSignature(mentions[0], mentions[-1]))
                    mentions = []

                mentions.extend(positions)
                fact_type = current_type
                continue

            # outside an entity

            if fact_type is not None:
                # finish current entity
                fact_signatures[fact_type].append(MentionSignature(mentions[0], mentions[-1]))

                fact_type = None
                mentions = []

        if fact_type is not None:
            # finish current entity
            fact_signatures[fact_type].append(MentionSignature(mentions[0], mentions[-1]))

        return fact_signatures

    def decode_label_ids(
            self,
            node: TIETextNode,
            label_ids: LongTensor,
            *,
            type_mapping: dict[str, set[MentionedFactsFactory]],
            label_ids_mapping: Sequence[str],
            precomputed_tokens: Sequence[Span] | None = None
    ) -> Iterator[AbstractFact]:

        fact_signatures = self.decode_signatures(label_ids, label_ids_mapping)
        tokens = precomputed_tokens if precomputed_tokens is not None else tuple(node.tokens)

        def build_facts_for_signatures(label: str, signatures: list[tuple[int, int]]) -> Iterator[AbstractFact]:
            for factory in type_mapping[label]:
                for start, end in signatures:  # convert signatures to mentions
                    mention = TextNodeMention(node, tokens[start].start, tokens[end].end)
                    yield from factory(mention, FactStatus.NEW)

        yield from chain.from_iterable(starmap(build_facts_for_signatures, fact_signatures.items()))

    @abstractmethod
    def _encode_entity(self, span_count: int, current_type: str, prev_label: str) -> list[str]:
        pass


class BIOLabellingStrategy(_BIOFactoryLabellingStrategy):
    def __init__(self):
        super().__init__('B', {'B', 'I'})

    def _encode_entity(self, span_count: int, current_type: str, prev_label: str) -> list[str]:
        return ['B'] + ['I'] * (span_count - 1)

    @property
    def type(self) -> LabellingStrategyType:
        return LabellingStrategyType.BIO


class IOBLabellingStrategy(_BIOFactoryLabellingStrategy):
    def __init__(self):
        super().__init__('B', {'B', 'I'})

    def _encode_entity(self, span_count: int, current_type: str, prev_label: str) -> list[str]:
        first = ['B'] if self._is_object(prev_label) and current_type == self._get_object_type(prev_label) else ['I']
        return first + ['I'] * (span_count - 1)

    @property
    def type(self) -> LabellingStrategyType:
        return LabellingStrategyType.IOB


_LABELLING_STRATEGIES = {
    LabellingStrategyType.BIO: BIOLabellingStrategy,
    LabellingStrategyType.IOB: IOBLabellingStrategy
}


def get_labelling_strategy(name: str | LabellingStrategyType) -> AbstractLabellingStrategy:
    strategy_type = LabellingStrategyType(name)
    return _LABELLING_STRATEGIES[strategy_type]()



# moved from tie_ml_base.helper

from dataclasses import dataclass, fields
from typing import Any, Sequence, TypeVar

import numpy as np
import torch
from torch import Tensor


def fill_roll_np(arr: np.ndarray, *, shift: int, fill: Any) -> np.ndarray:
    rolled = np.roll(arr, shift=shift)
    if shift > 0:
        rolled[:shift] = fill
    if shift < 0:
        rolled[shift:] = fill
    return rolled


def fill_roll(arr: Tensor, *, shift: int, fill: Any) -> Tensor:
    rolled = torch.roll(arr, shifts=shift)
    if shift > 0:
        rolled[:shift] = fill
    if shift < 0:
        rolled[shift:] = fill
    return rolled


def one_hot(arr: np.ndarray, *, num_classes: int, dtype: object = int) -> np.ndarray:
    incorrect_mask = ((arr >= num_classes) | (arr < 0))
    if incorrect_mask.any():
        incorrect_values = set(arr[incorrect_mask])
        raise ValueError(
            f'Cannot create one-hot vector (num_classes={num_classes}) from array with the following values present: {incorrect_values}!'
        )

    out = np.zeros((arr.size, num_classes), dtype=dtype)
    out[np.arange(arr.size), arr.ravel()] = 1
    out.shape = arr.shape + (num_classes,)
    return out


def pad_sequence(sequence: Sequence[np.ndarray], *, padding_value: Any) -> np.ndarray:
    padded_size = max(map(len, sequence))

    def padder(arr: np.ndarray) -> np.ndarray:
        pad_shape = [[0, 0] for _ in arr.shape]
        pad_shape[0][-1] = padded_size - len(arr)
        return np.pad(arr, pad_shape, 'constant', constant_values=padding_value)

    return np.stack(tuple(map(padder, sequence)))


def safe_concat(arrays: Sequence[np.ndarray], *, dtype: object) -> np.ndarray:
    if len(arrays):
        return np.concatenate(arrays)
    return np.empty(0, dtype=dtype)


@dataclass
class TupleLike:
    def __iter__(self):  # so it behaves like a NamedTuple
        return iter(self.as_tuple())

    def as_tuple(self):
        return tuple(self.__getattribute__(field.name) for field in fields(self))


_TensorStorage = TypeVar('_TensorStorage', bound='TensorStorage')


class TensorStorage:

    def to(self: _TensorStorage, device: torch.device) -> _TensorStorage:
        for field in fields(self):
            value = self.__getattribute__(field.name)
            if torch.is_tensor(value):
                object.__setattr__(self, field.name, value.to(device))
                # self.__setattr__(field.name, value.to(device))
        return self



from abc import ABCMeta, abstractmethod

from torch import Tensor

from tp_interfaces.serializable import Serializable


class AbstractPooler(Serializable, metaclass=ABCMeta):

    @abstractmethod
    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        pass


def choose(tensor: Tensor, from_idx: int, to_idx: int, *, dim: int) -> Tensor:
    return tensor.swapdims(0, dim)[from_idx:to_idx].swapdims(0, dim)



from enum import Enum
from functools import partial
from typing import Any, Callable, Type, Union

import torch
from torch import Tensor

from tie_ml_utils.torch_wrapper import DirectorySerializableTorchModule
from tp_interfaces.abstract import ModelTypeFactory
from .abstract import AbstractPooler


def _get_select_pooler() -> Type[AbstractPooler]:
    from .select import SelectPooler
    return SelectPooler


def _get_concat_pooler() -> Type[AbstractPooler]:
    from .concat import ConcatPooler
    return ConcatPooler


def _get_aggregate_pooler() -> Type[AbstractPooler]:
    from .aggregate import AggregatePooler
    return AggregatePooler


def _get_attn_pooler() -> Type[AbstractPooler]:
    from .attention import AttentionPooler
    return AttentionPooler


def _add_args(factory: Callable[[], Type[AbstractPooler]], **args) -> Callable[[], Callable[[], AbstractPooler]]:
    def factory_wrapper() -> Callable[[], AbstractPooler]:
        return partial(factory(), **args)

    return factory_wrapper


def _max_wrapper(tensor: Tensor, dim: int) -> Tensor:
    return torch.max(tensor, dim).values


class PoolerType(Enum):
    AGGREGATE = 'aggregate'
    CONCAT = 'concat'
    MAX = 'max'
    MEAN = 'mean'
    LAST = 'last'
    SELECT = 'select'
    SUM = 'sum'
    FIRST = 'first'
    ATTN = 'attention'


POOLERS = ModelTypeFactory({
    PoolerType.AGGREGATE.value: _get_aggregate_pooler,
    PoolerType.CONCAT.value: _get_concat_pooler,
    PoolerType.MAX.value: _add_args(_get_aggregate_pooler, aggregate_function=_max_wrapper),
    PoolerType.MEAN.value: _add_args(_get_aggregate_pooler, aggregate_function=torch.mean),
    PoolerType.LAST.value: _add_args(_get_select_pooler, index=-1),
    PoolerType.SELECT.value: _get_select_pooler,
    PoolerType.SUM.value: _add_args(_get_aggregate_pooler, aggregate_function=torch.sum),
    PoolerType.FIRST.value: _add_args(_get_select_pooler, index=0),
    PoolerType.ATTN.value: _get_attn_pooler
})


class NamedPooler(AbstractPooler, DirectorySerializableTorchModule):

    def __init__(self, name: str, pooler: AbstractPooler):
        DirectorySerializableTorchModule.__init__(self)
        self._pooler = pooler
        self._name = name

    @property
    def name(self) -> str:
        return self._name

    def forward(self, *args, **kwargs) -> Any:
        self.pool(*args, **kwargs)

    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        return self._pooler.pool(tensor, dim=dim)


def configure_pooler(config_or_pooler_type: Union[str, PoolerType, dict]) -> NamedPooler:
    if isinstance(config_or_pooler_type, str):
        config_or_pooler_type = PoolerType(config_or_pooler_type)
    if isinstance(config_or_pooler_type, PoolerType):
        model_name = config_or_pooler_type.value
        model_args = {}
    else:
        model_name = config_or_pooler_type['model']
        model_args = config_or_pooler_type['config']
    return NamedPooler(model_name, POOLERS[model_name](**model_args))



from setuptools import find_packages, setup

with open("README.md", "r", encoding="utf-8") as f:
    long_description = f.read()

with open("VERSION", "r", encoding="utf-8") as f:
    version = f.read()

setup(
    name='talisman-ie-ml-utils',
    version=version,
    description='Talisman-IE ML utilities',
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='ISPRAS Talisman NLP team',
    author_email='modis@ispras.ru',
    maintainer='Vladimir Mayorov',
    maintainer_email='vmayorov@ispras.ru',
    packages=find_packages(include=['tie_ml_utils', 'tie_ml_utils.*']),
    install_requires=[
        'talisman-interfaces>=0.8,<0.12',
        'talisman-ie-datamodel>=0.6.11,<0.7',
        'talisman-tools>=0.10,<0.12',
        'talisman-dm>=1.3.1,<2',
        'torch~=2.1',
        'numpy>=1.23.5,<3',
        'typing-extensions>=4.0.0',
        'pynvml~=11.5'
    ],
    entry_points={},
    data_files=[('', ['VERSION'])],
    python_requires='>=3.10',
    license='Apache Software License',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Intended Audience :: Developers',
        'Operating System :: OS Independent',
        'License :: OSI Approved :: Apache Software License'
    ]
)



# moved from tie_ml_base.env

import os
from typing import Optional

from tp_interfaces.helpers.convert import string_to_bool

DEFAULT_AMP_OPTIMIZE = False
DEFAULT_ASYNC_REQUESTS_SIZE = 16
DEFAULT_CPU_BATCH_SIZE = 1
DEFAULT_GPU_BATCH_SIZE = 1
DEFAULT_GPU_OOM_TIMEOUT = 30.0
DEFAULT_GPU_UTILIZATION_LIMIT = 1.0
DEFAULT_GPU_OOM_REPEAT_LIMIT = 3
DEFAULT_MAX_LENGTH = 40000


def get_bert_example_length() -> Optional[int]:
    bert_example_length_str = os.environ.get('BERT_EXAMPLE_LENGTH')
    return int(bert_example_length_str) if bert_example_length_str is not None else None


def get_bert_max_word_subtokens() -> Optional[int]:
    bert_max_word_subtokens = os.environ.get('BERT_MAX_WORD_SUBTOKENS')
    return int(bert_max_word_subtokens) if bert_max_word_subtokens is not None else None


def get_amp_optimize() -> bool:
    amp_optimize = os.environ.get('AMP_OPTIMIZE')
    return string_to_bool(amp_optimize) if amp_optimize is not None else DEFAULT_AMP_OPTIMIZE


def get_async_requests_size() -> int:
    async_requests_size_str = os.environ.get('ASYNC_REQUESTS_SIZE')
    return int(async_requests_size_str) if async_requests_size_str is not None else DEFAULT_ASYNC_REQUESTS_SIZE


def get_cpu_batch_size() -> int:
    cpu_batch_size_str = os.environ.get('CPU_BATCH_SIZE')
    return int(cpu_batch_size_str) if cpu_batch_size_str is not None else DEFAULT_CPU_BATCH_SIZE


def get_gpu_batch_size() -> int:
    gpu_batch_size_str = os.environ.get('GPU_BATCH_SIZE')
    return int(gpu_batch_size_str) if gpu_batch_size_str is not None else DEFAULT_GPU_BATCH_SIZE


def get_gpu_oom_timeout() -> float:
    gpu_oom_timeout = os.environ.get('GPU_OOM_TIMEOUT')
    return float(gpu_oom_timeout) if gpu_oom_timeout is not None else DEFAULT_GPU_OOM_TIMEOUT


def get_gpu_utilization_limit() -> float:
    gpu_utilization_limit = os.environ.get('GPU_UTILIZATION_LIMIT')
    return float(gpu_utilization_limit) if gpu_utilization_limit is not None else DEFAULT_GPU_UTILIZATION_LIMIT


def get_gpu_oom_repeat_limit() -> int:
    gpu_oom_repeat_limit = os.environ.get('GPU_OOM_REPEAT_LIMIT')
    return int(gpu_oom_repeat_limit) if gpu_oom_repeat_limit is not None else DEFAULT_GPU_OOM_REPEAT_LIMIT


def get_max_length() -> int:
    max_length = os.getenv('MAX_LENGTH')
    return int(max_length) if max_length else DEFAULT_MAX_LENGTH



from typing import Optional

import torch
from torch import Tensor

from tp_interfaces.serializable import DirectorySerializableMixin
from .abstract import AbstractPooler, choose


class ConcatPooler(AbstractPooler, DirectorySerializableMixin):

    def __init__(self, from_index: Optional[int] = None, to_index: Optional[int] = None, concat_along_index: int = -1):
        DirectorySerializableMixin.__init__(self)
        self._from_index = from_index
        self._to_index = to_index
        self._concat_along_dim = concat_along_index

    def pool(self, tensor: Tensor, *, dim: int) -> Tensor:
        selected_elements = choose(tensor, self._from_index, self._to_index, dim=dim)
        tensor_list = torch.unbind(selected_elements, dim)
        return torch.concat(tensor_list, dim=self._concat_along_dim)
